"""simulating.py simulator deed module
   simulates the motion of UUV or USV etc

"""
#print("module {0}".format(__name__))

import math
import time
import struct
from collections import deque
import random


from ....aid.sixing import *
from ....base.globaling import *
from ....aid.odicting import odict
from ....aid import aiding, navigating
from ....aid.navigating import DEGTORAD, RADTODEG
from ....base import doing

from ....base.consoling import getConsole
console = getConsole()

class SimulatorMotionUuv(doing.DoerLapse):
    """UUV motion simulator class
    """
    Ioinits=odict(
        group = 'simulator.motion.uuv',
        speed = 'state.speed', speedRate = 'state.speedRate',
        velocity = 'state.velocity',
        depth = 'state.depth', depthRate = 'state.depthRate',
        pitch = 'state.pitch', pitchRate = 'state.pitchRate',
        altitude = 'state.altitude',
        heading = 'state.heading', headingRate = 'state.headingRate',
        position = 'state.position', location = 'state.location',
        rpm = 'goal.rpm', stern = 'goal.stern', rudder = 'goal.rudder',
        current = 'scenario.current', bottom = 'scenario.bottom',
        onset = 'scenario.onset', origin = 'scenario.origin',
        parms = dict(rpmLimit = 1200.0, sternLimit = 20.0, rudderLimit = 20.0,
                     gs = 0.0022, gpr = -0.4, gpp = 0.0, gdb = -0.1, ghr = -0.4))

    def __init__(self, **kw):
        """Initialize instance.
           because of Center of buoyancy Center of gravity separation when pitched
           the relationship between sternPlane and pitch rate is usually nonlinear
           uses poor linear approx
           better would be to make pitch rate function of pitch and stern

        """
        #call super class method
        super(SimulatorMotionUuv,self).__init__(**kw)

        #used in reset to speed up processing
        self.ionames = dict( speed = None, speedRate = None, bottom = None,
                             pitch = None, pitchRate = None,
                             depth = None, depthRate = None, altitude = None,
                             heading = None, headingRate = None)


    def _initio(self, group, speed, speedRate, velocity,
                 depth, depthRate, pitch, pitchRate, altitude,
                 heading, headingRate, position, location,
                 rpm, stern, rudder, current, bottom, onset, origin, parms = None, **kw):
        """ Override since legacy interface

            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    rpmLimit = max rpm of motor
                    sternLimit = max absolute stern angle
                    rudderLimit = max abs rudder angle
                    gs = gain speed due to rpm where speed = gs * rpm
                    gpr = gain pitch rate due to stern where pitch rate = gpr * stern
                    gpp = gain pitch rate due to pitch where pitch rate = gpp * pitch (neg)
                          (cbcg restoring force)
                    gdb = gain depth rate due to buoyancy where depth rate = gdb (neg)
                    ghr = gain heading rate dut to rudder where headingRate = ghr * rudder

              group.elapsed = share copy of time lapse for logging

           speed = share path name output water speed m/s
           speedRate = share path name output water speedRate m/s^2

           velocity = share path name output velocity over ground vector (north east) m/s

           depth = share path name output depth meters positive down
           depthRate = share path name output depthRate meters/second positive down
           pitch = share path name output pitch degrees pitch up positive
           pitchRate = share path name output pitch degrees/second = delta pitch / lapse
           altitude = share path name output altitude off bottom

           heading = share path name output heading deg north =0 positive clock wise
           headingRate = share path name output headingRate deg/sec = delta heading / lapse
           position = share path name output position vector (north, east) m wrt origin on plane
           location = share path name output location vector (lat, lon) m wrt sphere flat earth

           rpm = share path name input rpm
           rudder = share path name input rudder
           stern = share path name input stern plane degrees positive down
           current = share path name input water current vector (north east) m/s
           bottom = share path name input simulated bottom depth of ocean
           onset = share path name input onset (start) position vector (north east)
           origin = share path name input origin (start) location  (lat lon)

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share
           .speed = ref to speed share
           .speedRate = ref to speedRate share
           .velocity = ref to velocity share
           .depth = ref to depth share
           .depthRate = ref to depthRate share
           .pitch = ref to pitch share
           .pitchRate = ref to pitchRate share
           .altitude = ref to altitude share
           .heading = ref to heading share
           .headingRate = ref to headingRate share
           .position = ref to position north east share
           .location = ref to location lat lon share

           .rpm = ref to rpm share
           .stern = ref to stern share
           .rudder = ref to rudder share
           .current = ref to current share
           .bottom = ref to bottom share
           .onset = ref to onset share
           .origin = ref to origin share
        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(rpmLimit = 1200.0, sternLimit = 20.0,rudderLimit = 20.0,
                         gs = 0.0020, gpr = -0.4, gpp = 0.0, gdb = -0.1, ghr = -0.4)
        self.parm.create(**parms)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs create share if not exist but force update of value
        self.speed = self.store.create(speed).update(value = 0.0)
        self.speedRate = self.store.create(speedRate).update(value = 0.0)

        self.velocity = self.store.create(velocity)
        self.velocity.update(north = 0.0)
        self.velocity.update(east = 0.0)

        self.depth = self.store.create(depth).update(value = 0.0)
        self.depthRate = self.store.create(depthRate).update(value = 0.0)
        self.pitch = self.store.create(pitch).update(value = 0.0)
        self.pitchRate = self.store.create(pitchRate).update(value = 0.0)
        self.altitude = self.store.create(altitude).update(value = 0.0)
        self.heading = self.store.create(heading).update(value = 0.0)
        self.headingRate = self.store.create(headingRate).update(value = 0.0)

        self.position = self.store.create(position)
        self.position.update(north = 0.0) #preserves order
        self.position.update(east = 0.0)

        self.location = self.store.create(location)
        self.location.update(lat = 0.0) #preserves order
        self.location.update(lon = 0.0)

        #inputs create share if not exists and create value if not exist
        self.rpm = self.store.create(rpm).create(value = 0.0)
        self.stern = self.store.create(stern).create(value = 0.0)
        self.rudder = self.store.create(rudder).create(value = 0.0)
        self.current = self.store.create(current)
        self.current.create(north = 0.0) #preserves field order
        self.current.create(east = 0.0)
        self.bottom = self.store.create(bottom).create(value = 50.0)
        self.onset = self.store.create(onset)
        self.onset.create(north = 0.0) #preserves order
        self.onset.create(east = 0.0)
        self.origin = self.store.create(origin)
        self.origin.create(lat = 0.0) #preserves order
        self.origin.create(lon = 0.0)

    def reset(self, **kwa):
        """Resets simulated motion state to passed in parameters
        """
        for key, value in kwa.iteritems():
            if key in self.ionames:
                getattr(self, key).value = value

    def restart(self):
        """Restart motion  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

        #need to init position here since onset may be changed at build time by init command
        if self.position.stamp is None: #only on first run
            #init position to onset where vehicle starts relative to 0,0 origin
            self.position.update(north = self.onset.data.north, east = self.onset.data.east)

        #need to init location here since origin may be changed at build time by init command
        if self.location.stamp is None: #only on first run
            #init location to origin lat lon
            self.location.update(lat = self.origin.data.lat, lon = self.origin.data.lon)

    def action(self, **kw):
        """Updates simulated motion state of vehicle
        """
        super(SimulatorMotionUuv,self).action(**kw) #lapse updated here

        self.elapsed.value = self.lapse #store lapse for logging

        if self.lapse <= 0.0: #only evaluate if lapse positive so rate calc good
            return

        rpm = self.rpm.value
        rpm = min(abs(self.parm.data.rpmLimit),max(0.0, rpm)) #limit rpm

        rudder = self.rudder.value
        rudderLimit = abs(self.parm.data.rudderLimit)
        rudder = min(rudderLimit,max(-rudderLimit,rudder)) #limit rudder

        stern = self.stern.value
        sternLimit = abs(self.parm.data.sternLimit)
        stern = min(sternLimit,max(-sternLimit, stern)) #limit stern

        speedLast = self.speed.value
        speed = self.parm.data.gs * rpm
        self.speed.value = speed

        speedRate = (speed - speedLast)/self.lapse
        self.speedRate.value = speedRate

        #slant distance traveled
        slant = speed * self.lapse

        #Pitch
        pitchLast = self.pitch.value
        #linear approximation
        #pitchrate a function of stern (gpr) and restoring force cbcg (gpp) as a function of pitch
        pitchRate = self.parm.data.gpr * stern + self.parm.data.gpp * pitchLast
        self.pitchRate.value = pitchRate

        pitch = pitchLast + self.lapse * pitchRate
        self.pitch.value = pitch

        #Depth
        depthLast = self.depth.value #need for horizontal component position change
        #vertical component of slant is depth change
        #positive pitch decreases depth also buoyancy term gdb that changes depth
        depth = depthLast - slant * math.sin(DEGTORAD * pitch) + self.parm.data.gdb * self.lapse
        #depth delta calculation assumes that if trying to change depth
        #but can't when on suface that horizontal component of motion
        #still needs to be reduced since vehicle slows down when pitched
        depthChange = depth - depthLast #virtual depth change
        depth = max(0.0,depth) #can't have negative depth
        self.depth.value = depth

        depthRate = (depth - depthLast)/self.lapse #actual depth rate of change
        self.depthRate.value = depthRate

        #altitude
        altitude = self.bottom.value - depth
        self.altitude.value = altitude

        #Heading
        headingRate = self.parm.data.ghr * rudder
        self.headingRate.value = headingRate

        headingLast = self.heading.value #need for position change
        heading = headingLast +  self.lapse * headingRate

        #Position
        #average heading over time lapse to calc position change must avg before wrap around
        headAvg = (heading + headingLast)/2.0
        #don't remember why want heading to be [0, 360] instead of [-180, 180]
        #if the latter we could use navigating.Wrap2()
        heading %= 360.0 #mod 360 to wrap aound
        if (heading < 0.0): #if negative make equiv positive
            heading += 360.0
        self.heading.value = heading

        #scale by horiz component of slant when changing depth
        horizontal = (slant**2 + depthChange**2)**0.5

        #heading angle convention  0 is north increasing clockwise
        #cartesian angle convention is 0 east increasing counter clockwise
        #cartesion angle = pi/2 - heading angle
        #cos(pi/2 - psi) = sin(psi)
        #sin(pi/2 - psi) = cos(psi)

        north = self.position.data.north
        east = self.position.data.east

        deltaNorth = horizontal * math.cos(DEGTORAD * headAvg) +\
            self.current.data.north * self.lapse
        deltaEast = horizontal  * math.sin(DEGTORAD * headAvg) +\
            self.current.data.east * self.lapse

        self.velocity.update(north = (deltaNorth / self.lapse),
                             east = (deltaEast / self.lapse))

        self.position.update(north = north + deltaNorth, east = east + deltaEast)

        lat, lon = navigating.SphereLLByDNDEToLL(self.location.data.lat,self.location.data.lon,deltaNorth,deltaEast)
        self.location.update(lat = lat, lon = lon)

    def _expose(self):
        """
           prints out motion state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        print("speed = %0.3f, rpm = %0.3f depth = %0.3f  pitch = %0.3f stern plane = %0.3f" %\
              (self.speed.value, self.rpm.value, self.depth.value, self.pitch.value, self.stern.value))
        print("heading = %0.3f rudder = %0.3f" %\
              (self.heading.value, self.rudder.value))
        print("vel north = %0.3f vel east = %0.3f, pos north = %0.3f pos east = %0.3f" %\
              (self.velocity.data.north, self.velocity.data.east,
               self.position.data.north, self.position.data.east))

class SimulatorMotionUsv(doing.DoerLapse):
    """USV motion simulator class
    """
    Ioinits=odict(
        group = 'simulator.motion.usv',
        speed = 'state.speed', speedRate = 'state.speedRate',
        velocity = 'state.velocity',
        heading = 'state.heading', headingRate = 'state.headingRate',
        position = 'state.position',
        rpm = 'goal.rpm', rudder = 'goal.rudder',
        current = 'scenario.current',
        onset = 'scenario.onset',
        parms = dict(rpmLimit = 3000.0,  rudderLimit = 20.0,
                     gs = 0.0025,  ghr = -0.25))

    def __init__(self,  **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorMotionUsv,self).__init__(**kw)

        #used in reset to speed up processing
        self.ionames = dict( speed = None, speedRate = None,
                             heading = None, headingRate = None)

    def _initio(self, group, speed, speedRate,  velocity,
                 heading, headingRate, position,
                 rpm, rudder, current, onset, parms = None, **kw):
        """ Override since legacy interface

            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    rpmLimit = max rpm of motor
                    rudderLimit = max abs rudder angle
                    gs = gain speed  where speed = gs * rpm
                    ghr = gain heading rate where headingRate = ghr * rudder

              group.elapsed = share copy of time lapse for logging

           speed = share path name output water speed m/s
           speedRate = share path name output water speedRate m/s^2

           velocity = share path name output velocity over ground vector (north east) m/s

           heading = share path name output heading deg north =0 positive clock wise
           headingRate = share path name output headingRate deg/sec = delta heading / lapse
           position = share path name output position vector (north, east) m wrt origin on plane

           rpm = share path name input rpm
           rudder = share path name input rudder
           current = share path name input water current vector (north east) m/s
           onset = share path name input onset (start) position vector (north east)

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share
           .speed = ref to speed share
           .speedRate = ref to speedRate share
           .velocity = ref to velocity share
           .heading = ref to heading share
           .headingRate = ref to headingRate share
           .position = ref to position north east share

           .rpm = ref to rpm share
           .rudder = ref to rudder share
           .current = ref to current share
           .onset = ref to onset share
        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(rpmLimit = 3000.0, rudderLimit = 20.0,
                         gs = 0.0025, ghr = -0.001)

        self.parm.create(**parms)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs create share if not exist but force update of value
        self.speed = self.store.create(speed).update(value = 0.0)
        self.speedRate = self.store.create(speedRate).update(value = 0.0)

        self.velocity = self.store.create(velocity)
        self.velocity.update(north = 0.0)
        self.velocity.update(east = 0.0)

        self.heading = self.store.create(heading).update(value = 0.0)
        self.headingRate = self.store.create(headingRate).update(value = 0.0)

        self.position = self.store.create(position)
        self.position.update(north = 0.0) #preserves order
        self.position.update(east = 0.0)

        #inputs create share if not exists and create value if not exist
        self.rpm = self.store.create(rpm).create(value = 0.0)
        self.rudder = self.store.create(rudder).create(value = 0.0)
        self.current = self.store.create(current)
        self.current.create(north = 0.0) #preserves field order
        self.current.create(east = 0.0)
        self.onset = self.store.create(onset)
        self.onset.create(north = 0.0) #preserves order
        self.onset.create(east = 0.0)
        #init onset position where vehicle starts relative to 0,0 origin
        self.position.update(north = self.onset.data.north, east = self.onset.data.east)

    def reset(self, **kwa):
        """Resets simulated motion state to passed in parameters
        """
        for key, value in kwa.iteritems():
            if key in self.ionames:
                getattr(self, key).value = value

    def restart(self):
        """Restart motion  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0
        #self.position.update(self.onset.items())

    def action(self, **kw):
        """Updates simulated motion state of vehicle
        """
        super(SimulatorMotionUsv,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        if self.lapse <= 0.0: #only evaluate controller if lapse positive so rate calc good
            return

        rpm = self.rpm.value
        rpm = min(abs(self.parm.data.rpmLimit),max(0.0, rpm)) #limit rpm

        rudder = self.rudder.value
        rudderLimit = abs(self.parm.data.rudderLimit)
        rudder = min(rudderLimit,max(-rudderLimit,rudder)) #limit rudder

        speedLast = self.speed.value
        speed = self.parm.data.gs * rpm
        self.speed.value = speed

        speedRate = (speed - speedLast)/self.lapse
        self.speedRate.value = speedRate

        #Heading
        headingRate = self.parm.data.ghr * speed * rudder
        self.headingRate.value = headingRate

        headingLast = self.heading.value #need for position change
        heading = headingLast +  self.lapse * headingRate

        #Position
        #average heading over time lapse to calc position change must avg before wrap around
        headAvg = (heading + headingLast)/2.0
        #don't remember why want heading to be [0, 360] instead of [-180, 180]
        #if the latter we could use navigating.Wrap2()
        heading %= 360.0 #mod 360 to wrap around
        if (heading < 0.0): #if negative make equiv positive
            heading += 360.0
        self.heading.value = heading

        #distance traveled
        horizontal = speed * self.lapse

        #heading angle convention  0 is north increasing clockwise
        #cartesian angle convention is 0 east increasing counter clockwise
        #cartesion angle = pi/2 - heading angle
        #cos(pi/2 - psi) = sin(psi)
        #sin(pi/2 - psi) = cos(psi)
        north = self.position.data.north
        east = self.position.data.east

        deltaNorth = horizontal * math.cos(DEGTORAD * headAvg) +\
            self.current.data.north * self.lapse
        deltaEast = horizontal  * math.sin(DEGTORAD * headAvg) +\
            self.current.data.east * self.lapse

        self.velocity.update(north = (deltaNorth / self.lapse),
                             east = (deltaEast / self.lapse))

        self.position.update(north = north + deltaNorth, east = east + deltaEast)

    def _expose(self):
        """
           prints out motion state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        print("speed = %0.3f, rpm = %0.3f heading = %0.3f rudder = %0.3f north = %0.3f east = %0.3f" %\
              (self.speed.value, self.rpm.value, self.heading.value, self.rudder.value))
        print("vel north = %0.3f vel east = %0.3f, pos north = %0.3f pos east = %0.3f" %\
              (self.velocity.data.north, self.velocity.data.east,
               self.position.data.north, self.position.data.east))

class SimulatorSensorGps(doing.DoerLapse):
    """GPS sensor simulator class
    """
    Ioinits=odict(
        group = 'simulator.sensor.gps',
        positionOut = 'gps.position', velocityOut = 'gps.velocity',
        error = 'gps.error',
        heading = 'heading.output', speed = 'state.speed',
        positionIn = 'state.position', velocityIn = 'state.velocity',
        scenario = 'scenario.gps',
        parms = dict(noiseBand = 5.0,  noiseJitter = 2.5,
                     noiseVelocity = 0.1))

    def __init__(self, **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorSensorGps,self).__init__(**kw)

    def _initio(self, group, positionOut, velocityOut, error,
                 heading, speed,  positionIn, velocityIn,
                 scenario, parms = None, **kw):
        """ Override since legacy interface

            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    noiseBand = meters max error in position from noise simulated
                    noiseJitter = meters/second max delta position noise
                    noiseVelocity = sigma of velocity noise

              group.elapsed = share copy of time lapse for logging

           positionOut = share path name output position meters north east GPS sim
           velocityOut = share path name output velocity m/s north east for GPS sim
           error = share path name output position error


           heading = share path name input heading deg north =0 positive clock wise
           speed = share path name input speed m/s

           positionIn = share path name input position vector (north, east) m wrt origin on plane
           velocityIn = share path name input ground velocity north east m/s

           scenario = share path name input scenario (for dropouts etc)
           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .positionOut = ref to gps position share
           .velocityOut = ref to gps velocity share
           .error = ref to gps position error


           .heading = ref to input heading share
           .speed = ref to input speed share
           .positionIn = ref to input position north east share
           .velocityIn = relf to input velocity share

           .scenario = ref to input scenario dropout

        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(noiseBand = 5.0, noiseJitter = 2.5, noiseVelocity = 0.1)

        self.parm.create(**parms)
        self.parm.data.noiseBand = abs(self.parm.data.noiseBand)
        self.parm.data.noiseJitter = abs(self.parm.data.noiseJitter)
        self.parm.data.noiseVelocity = abs(self.parm.data.noiseVelocity)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs create share if not exist but force update of value
        self.positionOut = self.store.create(positionOut)
        self.positionOut.update(north = 0.0) #preserves order
        self.positionOut.update(east = 0.0)

        self.velocityOut = self.store.create(velocityOut)
        self.velocityOut.update(north = 0.0) #preserves order
        self.velocityOut.update(east = 0.0)

        self.error = self.store.create(error)
        self.error.update(north = 0.0) #preserves order
        self.error.update(east = 0.0)

        #inputs create share if not exists and create value if not exist
        self.heading = self.store.create(heading).create(value = 0.0)
        self.speed = self.store.create(speed).create(value = 0.0)

        self.positionIn = self.store.create(positionIn)
        self.positionIn.create(north = 0.0)#preserves order
        self.positionIn.create(east = 0.0)

        self.velocityIn = self.store.create(velocityIn)
        self.velocityIn.create(north = 0.0)#preserves order
        self.velocityIn.create(east = 0.0)

        self.scenario = self.store.create(scenario).create(dropout = 0)
        #self.position.update(self.idealPosition.items()) #init position

    def restart(self):
        """Restart motion  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated motion state of vehicle
        """
        super(SimulatorSensorGps,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        if self.lapse <= 0.0: #only evaluate controller if lapse positive so rate calc good
            return

        if self.scenario.data.dropout: #location where no gps
            return

        #not used speed & heading
        speed = self.speed.value
        heading = self.heading.value

        en = self.error.data.north
        ee = self.error.data.east

        #jitter = self.parm.data.noiseJitter * self.lapse
        jitter = self.parm.data.noiseJitter
        band = self.parm.data.noiseBand
        vsigma = self.parm.data.noiseVelocity

        stamp = self.stamp

        en = min(band, max(-band, 0.6 * en + 0.4 * random.gauss(0.0, jitter)))
        #en = min(band, max(-band, 3 * jitter * math.cos(stamp * math.pi/2.0) + random.gauss(0.0, jitter/3.0)))
        #en = min(band, max(-band, 0.8 * en + 0.2 * random.gauss(0.0, jitter)))
        #en = min(band, max(-band, 0.9 * errorNorth + 0.1 * random.uniform(-jitter, jitter)))

        ee = min(band, max(-band, 0.6 * ee + 0.4 * random.gauss(0.0, jitter)))
        #ee = min(band, max(-band, 3 *jitter * math.cos(stamp * math.pi/2.0) + random.gauss(0.0, jitter/3.0)))
        #ee = min(band, max(-band, 0.8 * ee + 0.2 * random.gauss(0.0, jitter)))
        #ee = min(band, max(-band, 0.9 * errorEast + 0.1 * random.uniform(-jitter, jitter)))

        vsigma3 = 3.0 * vsigma
        evn = min(vsigma3, max(-vsigma3, random.gauss(0.0, vsigma)))
        eve = min(vsigma3, max(-vsigma3, random.gauss(0.0, vsigma)))

        self.positionOut.update(north = self.positionIn.data.north + en,
                                east = self.positionIn.data.east + ee)
        self.velocityOut.update(north = self.velocityIn.data.north + evn,
                                east = self.velocityIn.data.east + eve)
        self.error.update(north = en, east = ee)

    def _expose(self):
        """
           prints out sensor state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        print("north = %0.3f east = %0.3f, vel north = %0.3f vel east = %0.3f" %\
              (self.position.data.north, self.position.data.east,
               self.velocity.data.north, self.velocity.data.east))

class SimulatorSensorDvl(doing.DoerLapse):
    """DVLSensorSimulator DeedLapse Deed Class
       DVL sensor simulator class
    """
    Ioinits=odict(
        group = 'simulator.sensor.dvl',
        velocity = 'dvl.velocity', currentOut = 'dvl.current',
        altitude = 'dvl.altitude',
        heading = 'heading.output', speed = 'state.speed',
        currentIn = 'scenario.current',
        bottom = 'scenario.bottom',
        scenario = 'scenario.dvl',
        parms = dict(velSigma = 0.01, bias = 0.1, altSigma = 0.01))

    def __init__(self, **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorSensorDvl,self).__init__(**kw)

    def _initio(self, group, velocity, currentOut, altitude,
                 heading, speed, currentIn, bottom,
                 scenario, parms = None, **kw):
        """ Override since legacy interface

            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    velSigma = noise parameter  in velocity measurement
                    bias = bias parameter in velocity measurement
                    altSigma = noise parameter altitude measurements


              group.elapsed = share copy of time lapse for logging

           velocity = share path name output velocity mm/s forward starboard sim
           currentOut = share path name output current m/s forward starboard sim
           altitude = share path name output altitude m sim

           heading = share path name input heading deg north =0 positive clock wise
           speed = share path name input speed m/s

           currentIn = share path name input current vector (north, east) m wrt origin on plane
           bottom = share path name input bottom depth m
           scenario = share path name input scenario (for dropouts etc)
           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .velocity = ref to dvl velocity share
           .current = ref to dvl current share
           .altitude = ref to altitude share

           .heading = ref to heading share
           .speed = ref to speed share

           .currentIn = ref to input current north east share
           .bottom = ref to input bottom
           .scenario = ref to input dvl scenario for dropout
        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(velSigma = 0.01, bias = 0.001, altSigma = 0.01)

        self.parm.create(**parms)
        self.parm.data.velSigma = abs(self.parm.data.velSigma)
        self.parm.data.altSigma = abs(self.parm.data.altSigma)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs create share if not exist but force update of value

        self.velocity = self.store.create(velocity)
        self.velocity.update(forward = 0.0) #preserves order
        self.velocity.update(starboard = 0.0)

        self.currentOut = self.store.create(currentOut)
        self.currentOut.update(forward = 0.0) #preserves order
        self.currentOut.update(starboard = 0.0)
        self.currentOut.update(north = 0.0) #preserves order
        self.currentOut.update(east = 0.0)

        self.altitude = self.store.create(altitude).create(value = 0.0)

        #inputs create share if not exists and create value if not exist
        self.heading = self.store.create(heading).create(value = 0.0)
        self.speed = self.store.create(speed).create(value = 0.0)
        self.currentIn = self.store.create(currentIn)
        self.currentIn.create(north = 0.0)#preserves order
        self.currentIn.create(east = 0.0)
        self.bottom = self.store.create(bottom).create(value = 0.0)
        self.scenario = self.store.create(scenario).create(dropout = 0)


    def restart(self):
        """Restart motion  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated motion state of vehicle
        """
        super(SimulatorSensorDvl,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        if self.lapse <= 0.0: #only evaluate controller if lapse positive so rate calc good
            return

        if self.scenario.data.dropout: #location where no dvl
            return


        heading = self.heading.value #need for position change
        speed = self.speed.value

        cn = self.currentIn.data.north
        ce = self.currentIn.data.east
        alt = self.bottom.value

        vsigma = self.parm.data.velSigma
        bias = self.parm.data.bias
        asigma = self.parm.data.altSigma

        #transform to body forward starboard
        cf, cs = navigating.RotateNEToFS(heading, cn, ce)
        vf = speed + cf
        vs = cs

        vsigma3 = vsigma * 3.0
        noise = min(vsigma3, max(-vsigma3, random.gauss(0.0, vsigma)))
        #vf += noise + bias * self.lapse
        #vs += noise + bias * self.lapse

        vf += noise + bias
        vs += noise + bias

        noise = min(vsigma3, max(-vsigma3, random.gauss(0.0, vsigma)))
        cf += noise
        cs += noise
        cn, ce = navigating.RotateFSToNE(heading, cf, cs)

        asigma3 = asigma * 3.0
        noise = min(asigma3, max(-asigma3, random.gauss(0.0, asigma)))
        alt += noise

        self.velocity.update(forward = vf, starboard = vs)
        self.currentOut.update(forward = cf, starboard = cs, north = cn, east = ce)
        self.altitude.value = alt

    def _expose(self):
        """
           prints out sensor state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "vel forward = %0.3f vel starboard = %0.3f"
        format += "cur forward = %0.3f cur starboard = %0.3f altitude = %0.3f"
        print(format %\
              (self.velocity.data.forward, self.velocity.data.starboard,
               self.current.data.forward, self.current.data.starboard,
               self.altitude.value))

class SimulatorSensorCompass(doing.DoerLapse):
    """compass sensor simulator class
    """
    Ioinits = odict(
        group = 'simulator.sensor.compass',
        output = 'compass',
        input = 'state.heading', depth = 'state.depth',
        scenario = 'scenario.magnetic',
        parms = dict(phase = 24.0, amp = 1.0, sigma = 0.1))

    def __init__(self, **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorSensorCompass,self).__init__(**kw)

    def _initio(self, group, output,input, scenario, parms = None, **kw):
        """ Override since legacy interface

            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    phase = phase of sinusoid error component degrees
                    amp = amplitude of sinusoid error component degrees
                    sigma = noise stdev of error component degrees


              group.elapsed = share copy of time lapse for logging

           output = share path name output magnetic heading degrees

           input = share path name input heading true deg north =0 positive clock wise
              true = mag + declination
           scenario = share path name input scenario (for scenario.declination)
           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .output = ref to dvl velocity share
           .input = ref to heading share
           .scenario = ref to input scenario declination

        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(phase = 45, amp = 1.0, sigma = 0.1)

        self.parm.create(**parms)
        #phase can be negative
        self.parm.data.amp = abs(self.parm.data.amp)
        self.parm.data.sigma = abs(self.parm.data.sigma)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs
        self.output = self.store.create(output).update(value = 0.0)
        #inputs
        self.input = self.store.create(input).create(value = 0.0)
        self.scenario = self.store.create(scenario).create(declination = 0.0)

    def restart(self):
        """Restart motion  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated sensor
        """
        super(SimulatorSensorCompass,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        if self.lapse <= 0.0: #only evaluate controller if lapse positive so rate calc good
            return

        trueHeading = self.input.value #need for position change
        declination =  self.scenario.data.declination

        phase = self.parm.data.phase
        amp = self.parm.data.amp
        sigma = self.parm.data.sigma

        sigma3 = sigma * 3.0
        noise = min(sigma3, max(-sigma3, random.gauss(0.0, sigma)))

        error = amp * math.cos(DEGTORAD * (trueHeading + phase)) + noise

        heading = trueHeading - declination + error #magnetic heading with error
        heading %= 360.0 #mod 360 in case wrapped around
        if (heading < 0.0): #if negative make equiv positive
            heading += 360.0

        self.output.value = heading

    def _expose(self):
        """
           prints out sensor state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "heading = %0.3f phase = %0.3f amp = %0.3f sigma = %0.3f"
        print(format %\
              (self.output.value, self.parm.data.phase,
               self.parm.data.amp, self.parm.data.sigma))

class SimulatorSalinityLinear(doing.DoerLapse):
    """linear salinity simulator class
    """
    Ioinits = odict(
        group = 'simulator.salinity.linear',
        output = 'ctdsim', depth = 'state.depth',
        input = 'state.position',
        parms = dict(track = 0.0, north = 0.0, east = 0.0,
                     middle = 32.0, spread = 4.0, rising = True, width = 500.0,
                     layer = 20.0, shift = 2.0))

    def __init__(self, **kw):
        """Initialize instance.


           inherited instance attributes
           .stamp = time stamp
           .lapse = time lapse between updates of controller
           .name
           .store

        """
        #call super class method
        super(SimulatorSalinityLinear,self).__init__(**kw)


    def _initio(self, group, output,input, depth, parms = None, **kw):
        """ Override since legacy interface

                       group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    track = heading of gradient middle
                    north = north coord for gradient middle
                    east = east coord for gradient middle
                    middle = salinity at center of front
                    spread = total variation in salinity middle - spread/2 to middle + spread/2
                    rising = direction of gradient rising from right to left is True
                    width = distance over which total variation occurs
                                spread occurs over trackline - width/2 to trackline + width/2
                    layer = depth where center of gradient is at middle value
                    shift = shift in center of gradient with depth

              group.elapsed = share copy of time lapse for logging

           output = share path name output ctd

           input = share path name input position of vehicle

           depth = share path name depth of vehicle

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .output = ref to ctd salinity share
           .input = ref to vehicle position share
           .depth = ref to vehicle depth share
        """
        self.group = group

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs
        self.output = self.store.create(output).update(salinity = 0.0)

        #inputs
        self.input = self.store.create(input)
        self.input.create(north = 0.0)
        self.input.create(east = 0.0)
        self.depth = self.store.create(depth).create(value = 0.0)

        #parms
        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(track = 15, north = 0.0, east = 0.0, middle = 32.0,
                         spread  = 3.0, rising = True, width = 1000,
                         layer = 20.0, shift = 2.0)

        self.parm.create(**parms)

    def restart(self):
        """Restart sensor  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated sensor
        """
        super(SimulatorSalinityLinear,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        #if self.lapse <= 0.0: #only evaluate simulator if lapse positive so rate calc good
        #   return

        pn = self.input.data.north  #vehicle position
        pe = self.input.data.east

        track = self.parm.data.track
        north = self.parm.data.north #front center line formed from point and track
        east = self.parm.data.east
        middle = self.parm.data.middle #salinity at center of front
        spread = self.parm.data.spread
        rising = self.parm.data.rising #direction of gradient True = rising from right to left across center
        width = self.parm.data.width
        layer = self.parm.data.layer
        shift = self.parm.data.shift

        #generate distance of vehicle from center track
        d = navigating.DistancePointToTrack2D([east,north],track, [pe,pn])
        #shift center pos to right so shift d neg to left
        d -= (self.depth.value - layer) * shift
        d = max(- width/2.0, min( width/2.0, d)) #saturate

        if rising:  #sign of gradient rising from right to left
            s = 1.0
        else: #falling from right to left
            s = -1.0

        salinity = (s * d * 2.0/width) * spread/2.0 + middle

        self.output.update(salinity = salinity)

    def _expose(self):
        """
           prints out sensor state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "salinity = %0.3f track = %0.3f north = %0.3f east = %0.3f"
        format += " middle = %0.3f spread = %0.3f, rising = %s, width = %0.3f"
        format += " layer = %0.3f shift = %0.3f"
        print(format %\
              (self.output.salinity, self.parm.data.track,
               self.parm.data.north, self.parm.data.east, self.parm.data.middle,
               self.parm.data.spread, self.parm.data.rising, self.parm.data.width,
               self.parm.data.layer, self.parm.data.shift))

class SimulatorSalinitySinusoid(doing.DoerLapse):
    """salinity sensor simulator class
    """
    Ioinits = odict(
        group = 'simulator.salinity.sinusoid',
        output = 'ctdsim',
        input = 'state.position', depth = 'state.depth',
        parms = dict(track = 0.0, north = 0.0, east = 0.0,
                     middle = 32.0, spread = 4.0, rising = True, width = 500.0,
                     layer = 20.0, shift = 2.0))


    def __init__(self, **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorSalinitySinusoid,self).__init__(**kw)

    def _initio(self, group, output,input, parms = None, **kw):
        """ Override since legacy interface

                       group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    track = heading of gradient center
                    north = north coord for gradient center
                    eath = east coord for gradient center
                    middle = salinity at center of front
                    spread = total variation in salinity = middle +- spread
                    rising = direction of gradient rising from right to left is True
                    width = distance over which spread occurs track line +- width
                    amp = amplitude of sinusoid

              group.elapsed = share copy of time lapse for logging

           output = share path name output ctd

           input = share path name input position of vehicle

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .output = ref to ctd salinity share
           .input = ref to vehicle position share

        """
        self.group = group

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs
        self.output = self.store.create(output).update(salinity = 0.0)

        #inputs
        self.input = self.store.create(input).create(north = 0.0, east = 0.0)

        #parms
        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(track = 15, north = 0.0, east = 0.0, middle = 32.0,
                         spread  = 3.0, rising = True, width = 1000)

        self.parm.create(**parms)

    def restart(self):
        """Restart sensor  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated sensor
        """
        super(SimulatorSalinitySinusoid,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        #if self.lapse <= 0.0: #only evaluate simulator if lapse positive so rate calc good
        #   return

        pn = self.input.data.north  #vehicle position
        pe = self.input.data.east

        # center line (ray) of salinity front
        track = self.parm.data.track #direction = track
        north = self.parm.data.north #north coord of ref point base of ray
        east = self.parm.data.east #east coord of ref point base of ray

        middle = self.parm.data.middle #salinity at center of front
        spread = self.parm.data.spread #salinity range = middle +- spread
        #direction of gradient True = rising from right to left across center
        rising = self.parm.data.rising
        width = self.parm.data.width #spacial width of salinity front spread

        #generate signed distance of vehicle from center track (positive left)
        d = navigating.DistancePointToTrack2D([east,north],track, [pe,pn])

        d = max(- width/2.0, min( width/2.0, d)) #saturate

        if rising:  #sign of gradient rising from right to left
            s = 1.0
        else: #falling from right to left
            s = -1.0

        salinity = (s * d * 2.0/width) * spread + middle

        self.output.update(salinity = salinity)

    def _expose(self):
        """
           prints out sensor state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "salinity = %0.3f track = %0.3f north = %0.3f east = %0.3f"
        format += " middle = %0.3f spread = %0.3f, rising = %s, width = %0.3f"
        print(format %\
              (self.output.salinity, self.parm.data.track,
               self.parm.data.north, self.parm.data.east, self.parm.data.middle,
               self.parm.data.spread, self.parm.data.rising, self.parm.data.width))

class SimulatorGradient(doing.DoerLapse):
    """gradient simulator class
    """
    Ioinits = odict(
        group = 'simulator.gradient.depth',
        output = 'ctdsim', field = 'depth',
        position = 'state.position', depth = 'state.depth',
        parms = dict(track = 0.0, north = 0.0, east = 0.0,
                     middle = 32.0, spread = 4.0, rising = True, width = 500.0,
                     layer = 20.0, shift = 2.0, span = 10.0, height = 20.0, duct = 0))

    def __init__(self, **kw):
        """Initialize instance.
        """
        #call super class method
        super(SimulatorGradient,self).__init__(**kw)

    def _initio(self, group, output, field, position, depth, parms = None, **kw):
        """ Override since legacy interface

                       group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    track = heading of gradient middle
                    north = north coord for gradient middle
                    east = east coord for gradient middle
                    middle = salinity at center of front
                    spread = total horizontal variation in output middle - spread/2 to middle + spread/2
                    rising = direction of gradient rising from right to left is True
                    width = horisontal distance over which spread = total variation occurs
                                spread occurs over trackline - width/2 to trackline + width/2
                    layer = depth where center of gradient is at middle value
                    shift = shift in center of gradient with depth
                    span = total vertical variation
                    height = vertical distance over which span occurs to saturation
                    duct = type of vertical variation
                          -1 = cold at bottom, 0 = cold mid at layer, 1 = cold surface
              group.elapsed = share copy of time lapse for logging

           output = share path name output ctd
           field = field name of simulated output measurable
              also outputs depth field since filter needs depth

           position = share path name input position of vehicle

           depth = share path name input depth of vehicle

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .output = ref to ctd salinity share
           .position = ref to vehicle position share
           .depth = ref to vehicle depth share
        """
        self.group = group

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs
        self.field = field
        self.output = self.store.create(output)
        self.output.update({self.field : 0.0})
        self.output.update(depth = 0.0)

        #inputs
        self.position = self.store.create(position)
        self.position.create(north = 0.0)
        self.position.create(east = 0.0)
        self.depth = self.store.create(depth).create(value = 0.0)

        #parms
        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(track = 15, north = 0.0, east = 0.0, middle = 32.0,
                         spread  = 3.0, rising = True, width = 1000,
                         layer = 20.0, shift = 2.0, span = 10.0, height = 20.0, duct = 0 )

        self.parm.create(**parms)

    def restart(self):
        """Restart gradient  simulator

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

    def action(self, **kw):
        """Updates simulated sensor
        """
        super(SimulatorGradient,self).action(**kw) #lapse updated here

        #self.elapsed.updateJointly(value = self.lapse, stamp = stamp) #store lapse for logging
        self.elapsed.value = self.lapse

        #if self.lapse <= 0.0: #only evaluate simulator if lapse positive so rate calc good
        #   return

        pn = self.position.data.north  #vehicle position
        pe = self.position.data.east

        track = self.parm.data.track
        north = self.parm.data.north #front center line formed from point and track
        east = self.parm.data.east
        middle = self.parm.data.middle #salinity at center of front
        spread = self.parm.data.spread #horizontal variation of output
        rising = self.parm.data.rising #direction of gradient True = rising from right to left across center
        width = self.parm.data.width #horizontal distance over which spread occurs
        layer = self.parm.data.layer #depth of middle
        shift = self.parm.data.shift #horizontal shift of middle with depth
        span = self.parm.data.span #veritical variation of output
        height = self.parm.data.height #vertical distrance over which span occurs
        duct = self. parm.data.duct #Type of variation of output with depth in depth

        #horizontal distance of vehicle from center track
        dh = navigating.DistancePointToTrack2D([east,north],track, [pe,pn])
        #vertical distance from layer (pos below)
        dv = self.depth.value - layer

        #shift effective center of track as a function of depth
        dh -= dv * shift #shift center pos to right so shift d neg to left

        dh = max(- width/2.0, min( width/2.0, dh)) #saturate

        dv = max(- height/2.0, min( height/2.0, dv)) #saturate

        #compute temperature offset for duct type
        if duct == 0: #warmer as move away from later up and down
            dv = abs(dv)
        elif duct == -1: #colder down warmer up
            offset = -dv * span/float(height)
        elif duct == +1: #colder up warmer down
            offset = dv * span/float(height)
        else:
            offset = 0.0

        #sign of gradient rising = true then rising from right to left
        # else falling from right to left
        if not rising:
            dh = - dh

        out = middle + (dh * spread/float(width)) + (dv * span/float(height))

        self.output.update({self.field : out, 'depth' : self.depth.value})

    def _expose(self):
        """
           prints out gradient state

        """
        print("Simulator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "output = %0.3f track = %0.3f north = %0.3f east = %0.3f"
        format += " middle = %0.3f spread = %0.3f, rising = %s, width = %0.3f"
        format += " layer = %0.3f shift = %0.3f span = %0.3f height = %0.3f duct = %s"
        print(format %\
              (self.output[self.field], self.parm.data.track,
               self.parm.data.north, self.parm.data.east, self.parm.data.middle,
               self.parm.data.spread, self.parm.data.rising, self.parm.data.width,
               self.parm.data.layer, self.parm.data.shift, self.parm.data.span,
               self.parm.data.height, self.parm.data.duct))

SimulatorGradient.__register__('SimulatorGradientTemperature', ioinits=odict(
    group = 'simulator.gradient.temperature',
    output = 'ctdsim', field = 'temperature',
    position = 'state.position', depth = 'state.depth',
    parms = dict(track = 0.0, north = 0.0, east = 0.0,
                 middle = 32.0, spread = 4.0, rising = True, width = 500.0,
                 layer = 20.0, shift = 2.0, span = 10.0, height = 20.0, duct = 0)) )

SimulatorGradient.__register__('SimulatorGradientSalinity', ioinits=odict(
    group = 'simulator.gradient.salinity',
    output = 'ctdsim', field = 'salinity',
    position = 'state.position', depth = 'state.depth',
    parms = dict(track = 0.0, north = 0.0, east = 0.0,
                 middle = 32.0, spread = 4.0, rising = True, width = 500.0,
                 layer = 20.0, shift = 2.0, span = 10.0, height = 20.0, duct = 0)) )




"""
Python 2 to 3 supporting definitions

"""
from __future__ import absolute_import, division, print_function

import sys

# Python2to3 support
if sys.version > '3':
    long = int
    basestring = (str, bytes)
    unicode = str
    xrange = range

    def ns2b(x):
        """
        Converts from native str type to native bytes type
        """
        return x.encode('ISO-8859-1')

    def ns2u(x):
        """
        Converts from native str type to native unicode type
        """
        return x

    def reraise(kind, value, trace=None):
        if value is None:
            value = kind()
        if value.__traceback__ is not trace:
            raise value.with_traceback(trace)
        raise value



else:
    #long = long
    #basestring = basestring
    #unicode = unicode

    def ns2b(x):
        """
        Converts from native str type to native bytes type
        """
        return x

    def ns2u(x):
        """
        Converts from native str type to native unicode type
        """
        return unicode(x)

    exec("""def reraise(kind, value, trace=None): raise kind, value, trace""", globals())



"""
setup.py

Basic setup file to enable pip install
See:
    https://pythonhosted.org/setuptools/
    https://bitbucket.org/pypa/setuptools


python setup.py register sdist upload

"""
import sys
import os
from setuptools import setup, find_packages

# Change to Ioflo's source directory prior to running any command
try:
    SETUP_DIRNAME = os.path.dirname(__file__)
except NameError:
    # We're probably being frozen, and __file__ triggered this NameError
    # Work around this
    SETUP_DIRNAME = os.path.dirname(sys.argv[0])

if SETUP_DIRNAME != '':
    os.chdir(SETUP_DIRNAME)

SETUP_DIRNAME = os.path.abspath(SETUP_DIRNAME)

IOFLO_METADATA = os.path.join(SETUP_DIRNAME, 'ioflo', '__metadata__.py')

# Load the metadata using exec() so we don't trigger an import of ioflo.__init__
# This is mainly a problem for Python 2.6

exec(compile(open(IOFLO_METADATA).read(), IOFLO_METADATA, 'exec'))


PYTHON26_REQUIRES = []
if sys.version_info < (2, 7): #tuple comparison element by element
    PYTHON26_REQUIRES.extend(['importlib>=1.0.3',
                              'argparse>=1.2.1'])

REQUIRES = [] + PYTHON26_REQUIRES

if sys.version_info > (3,):
    PYTHON_SCRIPTS = ['scripts/ioflo', 'scripts/ioflo3',]
else:
    PYTHON_SCRIPTS = ['scripts/ioflo', 'scripts/ioflo2',]

setup(
    name='ioflo',
    version=__version__,
    description='Flow Based Programming Automated Reasoning Engine and Automation Operation System',
    long_description='Enabling the Programmable World. http://ioflo.com  ',
    url='https://github.com/ioflo/ioflo',
    download_url='https://github.com/ioflo/ioflo/archive/master.zip',
    author=__author__,
    author_email='info@ioflo.com',
    license=__license__,
    keywords=('Automation Operating System Automated Reasoning Engine '
              'Flow Based Programming Intelligent Automation Pub/Sub ioflo FloScript'),
    packages=find_packages(exclude=['test', 'test.*',
                                      'docs', 'docs*',
                                      'log', 'log*', 'ioflo/app/log*']),
    package_data={
        '':       ['*.txt',  '*.md', '*.rst', '*.json', '*.conf', '*.html',
                   '*.css', '*.ico', '*.png', 'LICENSE', 'LEGAL'],
        'ioflo': ['app/plan/*.flo', 'app/plan/*/*.flo',
                  'app/plan/*.txt', 'app/plan/*/*.txt',],},
    install_requires=REQUIRES,
    extras_require={},
    scripts=PYTHON_SCRIPTS,)




"""poking.py goal action module


"""
#print("module {0}".format(__name__))

import time
import struct

from collections import deque
try:
    from itertools import izip
except ImportError: # python3
    izip = zip
import inspect


from ..aid.sixing import *
from .globaling import *
from ..aid.odicting import odict

from ..aid import aiding
from . import excepting
from . import registering

from . import storing
from . import acting

from .consoling import getConsole
console = getConsole()

class Poke(acting.Actor):
    """Poke Class to put values into explicit shares"""
    Registry = odict()


class PokeDirect(Poke):
    """Class to put direct data values into destination share"""

    def _resolve(self, sourceData, destination, destinationFields, **kwa):
        parms = super(PokeDirect, self)._resolve( **kwa)

        destination = self._resolvePath(ipath=destination,  warn=True) # now a share
        srcFields = sourceData.keys()
        dstFields = self._prepareDstFields(srcFields, destination, destinationFields)

        dstData = odict()
        for dstField, srcField in izip(dstFields, srcFields):
            dstData[dstField] = sourceData[srcField]

        parms['data'] = dstData
        parms['destination'] = destination
        return parms

    def action(self, data, destination, **kwa):
        """ Put data into share
            parameters:
              data = data to copy from
              destination = share to copy to
        """
        console.profuse("Put {0} into {1}\n".format( data, destination.name))

        destination.update(data)

class PokeIndirect(Poke):
    """
    Class to copy values from one share to another
       based on source and destination field lists

    """
    def _resolve(self, source, sourceFields, destination, destinationFields, **kwa):
        parms = super(PokeIndirect, self)._resolve( **kwa)

        source = self._resolvePath(ipath=source,  warn=True) # now a share
        destination = self._resolvePath(ipath=destination,  warn=True) # now a share

        sourceFields, destinationFields = self._prepareSrcDstFields(source,
                                                        sourceFields,
                                                        destination,
                                                        destinationFields)

        parms['source'] = source
        parms['sourceFields'] = sourceFields
        parms['destination'] = destination
        parms['destinationFields'] = destinationFields
        return parms

    def action(self, source, sourceFields, destination, destinationFields, **kwa):
        """ Copy sourceFields in source to destinationFields in destination

            copy fields in order according to field lists
              field list order is significant
                 a field of same name in source and destination will
                 not be copied to each other unless appear in same place
                 in both field lists

            parameters:
                source = share to copy from
                sourceFields = list of fields to copy from
                destination = share to copy to
                destinationFields = list of fields to copy to

        """
        console.profuse("Copy {0} in {1} into {2} in {3}\n".format(
            sourceFields, source.name, destinationFields, destination.name))

        data = odict()

        for df, sf in izip(destinationFields, sourceFields):
            data[df] = source[sf]

        destination.update(data) #updates time stamp as well

        console.profuse("Copied {0} into {1}\n".format(
                    data, destination.name))
        return None

class IncDirect(Poke):
    """Class to incremate destination share by direct data values"""

    def _resolve(self, destination, destinationFields, sourceData, **kwa):
        parms = super(IncDirect, self)._resolve( **kwa)

        destination = self._resolvePath(ipath=destination,  warn=True) # now a share
        sourceFields = sourceData.keys()
        destinationFields = self._prepareDstFields(sourceFields,
                                          destination,
                                          destinationFields)

        dstData = odict()
        for dstField, srcField in izip(destinationFields, sourceFields):
            dstData[dstField] = sourceData[srcField]

        parms['destination'] = destination
        parms['data'] = dstData
        return parms

    def action(self, destination, data, **kwa):
        """ Increment corresponding items in destination by items in data

            if only one field then single increment
            if multiple fields then vector increment

            parameters:
                destination = share to increment
                sourceData = dict of field values to increment by
        """
        try:
            dstData = odict()
            for field in data:
                dstData[field] = destination[field] + data[field]
            destination.update(dstData) #update so time stamp updated, use dict
        except TypeError as ex: #in case value is not a number
            console.terse("Error in Inc: {0}\n".format(ex))
        else:
            console.profuse("Inc {0} in {1} by {2} to {3}\n".format(
                data.keys(), destination.name, data.values(), dstData.values()))


class IncIndirect(Poke):
    """
    Class to increment values in one share with values from another share
       based on source and destination field lists

    """
    def _resolve(self, destination, destinationFields, source, sourceFields, **kwa):
        parms = super(IncIndirect, self)._resolve( **kwa)

        destination = self._resolvePath(ipath=destination,  warn=True) # now a share
        source = self._resolvePath(ipath=source,  warn=True) # now a share

        sourceFields, destinationFields = self._prepareSrcDstFields(source,
                                                        sourceFields,
                                                        destination,
                                                        destinationFields)

        parms['destination'] = destination
        parms['destinationFields'] = destinationFields
        parms['source'] = source
        parms['sourceFields'] = sourceFields
        return parms

    def action(self, destination, destinationFields, source, sourceFields, **kwa):
        """ Increment destinationFields in destination by sourceFields in source
            parameters:
                destination = share to increment
                destinationField = field in share to increment
                source = share with value to increment by
                sourceField = field in share with value to increment by

        """
        try:
            data = odict()
            for dstField, srcField in izip(destinationFields, sourceFields):
                data[dstField] = destination[dstField] + source[srcField]
            destination.update(data) #update so time stamp updated, use dict
        except TypeError as ex:
            console.terse("Error in Inc: {0}\n".format(ex1))
        else:
            console.profuse("Inc {0} in {1} from {2} in {3} to {4}\n".format(
                destinationFields, destination.name, sourceFields, source.name, data.values))



"""estimating.py estimator deed module


"""
#print("module {0}".format(__name__))

import math
import time
import struct
from collections import deque
import inspect


from ....aid.sixing import *
from ....aid.odicting import odict
from ....base.globaling import *

from ....aid import aiding, navigating, blending
from ....aid.navigating import DEGTORAD, RADTODEG
from ....base import doing

from ....base.consoling import getConsole
console = getConsole()

class EstimatorPositionNfl(doing.DoerLapse):
    """Estimator Position NonlinearFusion class
    """
    Ioinits = odict(
        group = 'estimator.position.nlf',
        position = 'nlf.position',
        drPosition = 'dr.position', drBias = 'dr.bias',
        speed = 'state.speed', heading = 'heading.output',
        current = 'dvl.current',
        dvlVelocity = 'dvl.velocity',
        gpsPosition = 'gps.position', gpsVelocity = 'gps.velocity',
        parms = dict(upsilon = 5.0, scale = 2.0, gain = 0.01,
                     dvlStamp = 0.0, stale = 1.0,
                     gpsPosStamp = 0.0, gpsVelStamp = 0.0))

    def __init__(self, **kw):
        """Initialize instance.

           inherited instance attributes
           .stamp = time stamp
           .lapse = time lapse between updates of controller
           .name
           .store

        """
        #call super class method
        super(EstimatorPositionNfl,self).__init__(**kw)

    def _initio(self, group, position, drPosition, drBias,
                 speed, heading, current, dvlVelocity, gpsPosition, gpsVelocity,
                 parms = None, **kw):
        """ Override since uses legacy interface
            group is path name of group in store, group has following subgroups or shares:
              group.parm = share for data structure of fixed parameters or coefficients
                 parm has the following fields:
                    upsilon = fusion parameter related to uncertainty
                    scale = fusion scale factor
                    gain = bias filter gain
                    dvlStamp = time stamp last dvl update
                    stale = time delay for stale pos estimate
                    gpsPosStamp = time stamp last gps position update
                    gpsVelStamp = time stamp last gps velocity update

              group.elapsed = share copy of time lapse for logging

           position = share path name output fused position north east m
           drPosition = share path name output dead reckoned position
           drBias = share path name output dead reckoned forward starboard bias (drift rate m/s)

           speed = share path name input speed m/s
           heading = share path name input heading deg north =0 positive clock wise
           current = share path name input current vector (north, east) m wrt origin on plane
           dvlVelocity = share path name input dvl velocity forward starboard
           gpsPosition = share path name input gps position north sourth
           gpsVelocity = share path name input gps velocity

           parms = dictionary to initialize group.parm fields

           instance attributes

           .group = copy of group name
           .parm = reference to input parameter share

           .position = ref to output position share
           .drPosition = ref to output dr position share
           .drBias = ref to output dr bias share

           .speed = ref to speed share
           .heading = ref to heading share
           .current = ref to input current north east share
           .dvlVelocity = ref to input dvl velocity
           .gpsPosition = ref to input gps position
           .gpsVelocity = ref to input gps velocity


        """
        self.group = group

        self.parm = self.store.create(group + '.parm')#create if not exist
        if not parms:
            parms = dict(upsilon = 5.0, scale = 2.0, gain = 0.01,
                         dvlStamp = 0.0, stale = 1.0,
                         gpsPosStamp = 0.0, gpsVelStamp = 0.0)

        self.parm.create(**parms)
        self.parm.data.upsilon = abs(self.parm.data.upsilon)
        self.parm.data.scale = abs(self.parm.data.scale)
        self.parm.data.gain = abs(self.parm.data.gain)
        self.parm.data.stale = abs(self.parm.data.stale)

        self.elapsed = self.store.create(group + '.elapsed').create(value = 0.0)#create if not exist

        #outputs create share if not exist but force update of value

        self.position = self.store.create(position)
        self.position.create(north = 0.0).update(north = 0.0) #preserves order
        self.position.create(east = 0.0).update(east = 0.0)

        self.drPosition = self.store.create(drPosition)
        self.drPosition.update(north = 0.0) #preserves order
        self.drPosition.update(east = 0.0)

        self.drBias = self.store.create(drBias)
        self.drBias.update(forward = 0.0) #preserves order
        self.drBias.update(starboard = 0.0)

        #inputs create share if not exists and create value if not exist
        self.speed = self.store.create(speed).create(value = 0.0)
        self.heading = self.store.create(heading).create(value = 0.0)
        self.current = self.store.create(current)
        self.current.create(north = 0.0)#preserves order
        self.current.create(east = 0.0)

        self.dvlVelocity = self.store.create(dvlVelocity)
        self.dvlVelocity.create(forward = 0.0)#preserves order
        self.dvlVelocity.create(starboard = 0.0)

        self.gpsPosition = self.store.create(gpsPosition)
        self.gpsPosition.create(north = 0.0)#preserves order
        self.gpsPosition.create(east = 0.0)

        self.gpsVelocity = self.store.create(gpsVelocity)
        self.gpsVelocity.create(north = 0.0)#preserves order
        self.gpsVelocity.create(east = 0.0)

    def restart(self):
        """Restart

        """
        self.stamp = self.store.stamp
        self.lapse = 0.0

        self.parm.data.dvlStamp = self.dvlVelocity.stamp
        self.parm.data.gpsPosStamp = self.gpsPosition.stamp
        self.parm.data.gpsVelStamp = self.gpsVelocity.stamp

    def action(self, **kw):
        """Updates estimater
        """
        super(EstimatorPositionNfl,self).action(**kw) #.lapse & .stamp updated here
        #.lapse is time since last run
        #.stamp is current time

        self.elapsed.value = self.lapse #store lapse for logging

        #check here because rate calc divide by lapse
        #only finish update if lapse positive
        if self.lapse <= 0.0: #lapse non positive so return
            return

        #Two DR estimates: one uses DVL, other uses HSC (heading water speed current)
        #we need a DR estimate any time we get a GPS to make correction.
        #If we don't get new DVL data at same time that we get new gps then must use HSC
        #Also we don't want to go too long between DR estimates even if we
        #Don't get any DVL or GPS so we use HSC if last time since either new DVL
        #or new GPS is too long (stale)

        #generate events: newDVL, newGPS, newHSC
        staleNLF = False
        newDVL = False
        newGPS = False
        newHSC = False

        if (self.dvlVelocity.stamp > self.parm.data.dvlStamp):
            newDVL = True  #new dvl data since last time we used it

        if (self.gpsPosition.stamp > self.parm.data.gpsPosStamp):
            newGPS = True #new gps fix since last time we used it

        dvlAge = self.stamp - self.parm.data.dvlStamp
        gpsAge = self.stamp - self.parm.data.gpsPosStamp

        if (min(dvlAge, gpsAge) >= self.parm.data.stale):
            staleNLF = True #dvl and gps data are stale

        if not newDVL and (newGPS or staleNLF):
            newHSC = True #need to compute hsc

        if not (newDVL or newGPS or newHSC): #nothing to do so return
            return

        #compute DR displacement
        heading = self.heading.value #need for position change
        #get time lapse since last nlf position update
        nlfLapse = self.stamp - self.position.stamp

        if newDVL:
            #use nlfLapse here because we may update position even when no dvl update
            dvlForward = self.dvlVelocity.data.forward * nlfLapse #make distance
            dvlStarboard = self.dvlVelocity.data.starboard * nlfLapse #make distance
            nDisp, eDisp = navigating.RotateFSToNE(heading, dvlForward, dvlStarboard)

            self.parm.data.dvlStamp = self.stamp #update dvl last used Stamp

        else: #newHSC  we wouldn't be here unless we needed to update
            #use nlfLapse here because we may update position even when no hsc update
            speed = self.speed.value
            cn = self.current.data.north
            ce = self.current.data.east

            nDisp = (speed * math.cos(DEGTORAD * heading) + cn) * nlfLapse
            eDisp = (speed * math.sin(DEGTORAD * heading) + ce) * nlfLapse

        #compute positions (nlf and pure dr)
        #get last nlf position
        north = self.position.data.north
        east = self.position.data.east
        #compute dr based nlf position with dr displacement
        north += nDisp
        east += eDisp

        #get last pure deadreckoned position
        drNorth = self.drPosition.data.north
        drEast = self.drPosition.data.east
        #Update pure deadreckoned estimate
        drNorth += nDisp
        drEast += eDisp
        self.drPosition.update(north = drNorth, east = drEast)

        #best nlf estimate so far is (north, east)
        if newGPS: #new gps fix so make correction to (north, east)
            gpsNorth = self.gpsPosition.data.north
            gpsEast = self.gpsPosition.data.east

            nError = gpsNorth - north
            eError = gpsEast - east

            upsilon = self.parm.data.upsilon
            scale = self.parm.data.scale

            nDelta = nError * blending.blend1(nError, upsilon, scale)
            eDelta = eError * blending.blend1(eError, upsilon, scale)

            north += nDelta #make correction to dr disp
            east += eDelta #make correction to dr disp

            #update bias estimate
            #get old bias
            fBias = self.drBias.data.forward
            sBias = self.drBias.data.starboard

            gpsLapse = self.stamp - self.parm.data.gpsPosStamp

            fDelta, sDelta = navigating.RotateNEToFS(heading, nDelta, eDelta)
            gain = self.parm.data.gain

            fBias = (1.0 - gain) * fBias + gain * fDelta/gpsLapse
            sBias = (1.0 - gain) * sBias + gain * sDelta/gpsLapse
            self.drBias.update(forward = fBias, starboard = sBias)

            self.parm.data.gpsPosStamp = self.stamp #update gps used last stamp

        #update nlf position using best estimate (north, east)
        self.position.update(north = north, east = east)


    def _expose(self):
        """
           prints out sensor state

        """
        print("Estimator %s stamp = %s  lapse = %0.3f" % (self.name, self.stamp,self.lapse))
        format = "north = %0.3f east = %0.3f"
        print(format %\
              (self.position.data.north, self.position.data.east))



"""
httping.py

nonblocking http classes


"""
from __future__ import absolute_import, division, print_function


import sys
import os
import socket
import errno
import io
from collections import deque
import codecs
import json
import ssl
import copy
import random
import datetime

if sys.version > '3':
    from urllib.parse import urlsplit, quote, quote_plus, unquote, unquote_plus
else:
    from urlparse import urlsplit
    from urllib import quote, quote_plus, unquote, unquote_plus

try:
    import simplejson as json
except ImportError:
    import json

from email.parser import HeaderParser

# Import ioflo libs
from ..sixing import *
from ..odicting import odict, lodict, modict
from ...base import excepting, storing
from .. import aiding
from ..nonblocking import Outgoer, OutgoerTls, Server, ServerTls

from . import httping

from ...base.consoling import getConsole
console = getConsole()

CRLF = b"\r\n"
LF = b"\n"
CR = b"\r"


#  Class Definitions

class Requestant(httping.Parsent):
    """
    Nonblocking HTTP Server Requestant class
    Parses request msg
    """

    def __init__(self, incomer=None, **kwa):
        """
        Initialize Instance
        Parameters:
            incomer = Incomer connection instance

        """
        super(Requestant, self).__init__(**kwa)
        self.incomer = incomer
        self.url = u''   # full path in request line either relative or absolute
        self.scheme = u''  # scheme used in request line path
        self.hostname = u''  # hostname used in request line path
        self.port = u''  # port used in request line path
        self.path = u''  # partial path in request line without scheme host port query fragment
        self.query = u'' # query string from full path
        self.fragment = u''  # fragment from full path
        self.changed = False  # True if msg changed on last parse iteration

    def checkPersisted(self):
        """
        Checks headers to determine if connection should be kept open until
        client closes it
        Sets the .persisted flag
        """
        connection = self.headers.get("connection")  # check connection header
        if self.version == (1, 1):  # rules for http v1.1
            self.persisted = True  # connections default to persisted
            connection = self.headers.get("connection")
            if connection and "close" in connection.lower():
                self.persisted = False  # unless connection set to close.

            # non-chunked but persistent connections should have non None for
            # content-length Otherwise assume not persisted
            elif (not self.chunked and self.length is None):
                self.persisted = False

        elif self.version == (1, 0):  # rules for http v1.0
            self.persisted = False  # connections default to non-persisted
            # HTTP/1.0  Connection: keep-alive indicates persistent connection.
            if connection and "keep-alive" in connection.lower():
                self.persisted = True

    def parseHead(self):
        """
        Generator to parse headers in heading of .msg
        Yields None if more to parse
        Yields True if done parsing
        """
        if self.headed:
            return  # already parsed the head

        self.headers = lodict()

        # create generator
        lineParser = httping.parseLine(raw=self.msg, eols=(CRLF, LF), kind="status line")
        while True:  # parse until we get full start line
            if self.closed:  # connection closed prematurely
                raise httping.PrematureClosure("Connection closed unexpectedly "
                                               "while parsing request start line")

            line = next(lineParser)
            if line is None:
                (yield None)
                continue
            lineParser.close()  # close generator
            break

        method, url, version = httping.parseRequestLine(line)

        self.method = method
        self.url = url.strip()

        if not version.startswith(u"HTTP/1."):
            raise httping.UnknownProtocol(version)

        if version.startswith(u"HTTP/1.0"):
            self.version = (1, 0)
        else:
            self.version = (1, 1)  # use HTTP/1.1 code for HTTP/1.x where x>=1


        pathSplits = urlsplit(self.url)
        self.path = unquote(pathSplits.path)  # unquote non query path portion here
        self.scheme = pathSplits.scheme
        self.hostname = pathSplits.hostname
        self.port = pathSplits.port
        self.query = httping.unquoteQuery(pathSplits.query)  # unquote only the values
        self.fragment = pathSplits.fragment

        leaderParser = httping.parseLeader(raw=self.msg,
                                   eols=(CRLF, LF),
                                   kind="leader header line")
        while True:
            if self.closed:  # connection closed prematurely
                raise httping.PrematureClosure("Connection closed unexpectedly "
                                               "while parsing request header")

            headers = next(leaderParser)
            if headers is not None:
                leaderParser.close()
                break
            (yield None)
        self.headers.update(headers)

        # are we using the chunked-style of transfer encoding?
        transferEncoding = self.headers.get("transfer-encoding")
        if transferEncoding and transferEncoding.lower() == "chunked":
            self.chunked = True
        else:
            self.chunked = False

        # NOTE: RFC 2616, S4.4, #3 says ignore if transfer-encoding is "chunked"
        contentLength = self.headers.get("content-length")
        if contentLength and not self.chunked:
            try:
                self.length = int(contentLength)
            except ValueError:
                self.length = None
            else:
                if self.length < 0:  # ignore nonsensical negative lengths
                    self.length = None
        else:
            self.length = None

        contentType = self.headers.get("content-type")
        if contentType:
            if u';' in contentType: # should also parse out charset for decoding
                contentType, sep, encoding = contentType.rpartition(u';')
                if encoding:
                    self.encoding = encoding

            if 'application/json' in contentType.lower():
                self.jsoned = True
            else:
                self.jsoned = False

        # Should connection be kept open until client closes
        self.checkPersisted()  # sets .persisted

        self.headed = True
        yield True
        return

    def parseBody(self):
        """
        Parse body
        """
        if self.bodied:
            return  # already parsed the body

        if self.length and self.length < 0:
            raise ValueError("Invalid content length of {0}".format(self.length))

        del self.body[:]  # self.body.clear() clear body python2 bytearrays don't clear

        if self.chunked:  # chunked takes precedence over length
            self.parms = odict()
            while True:  # parse all chunks here
                if self.closed:  # connection closed prematurely
                    raise httping.PrematureClosure("Connection closed unexpectedly"
                                                   " while parsing request body chunk")

                chunkParser = httping.parseChunk(raw=self.msg)
                while True:  # parse another chunk
                    result = next(chunkParser)
                    if result is not None:
                        chunkParser.close()
                        break
                    (yield None)

                size, parms, trails, chunk = result

                if parms:  # chunk extension parms
                    self.parms.update(parms)

                if size:  # size non zero so append chunk but keep iterating
                    self.body.extend(chunk)

                    if self.closed:  # no more data so finish
                        chunkParser.close()
                        break

                else:  # last chunk when empty chunk so done
                    if trails:
                        self.trails = trails
                    chunkParser.close()
                    break

        elif self.length != None:  # known content length
            while len(self.msg) < self.length:
                if self.closed:  # connection closed prematurely
                    raise httping.PrematureClosure("Connection closed unexpectedly"
                                                   " while parsing request body")

                (yield None)

            self.body = self.msg[:self.length]
            del self.msg[:self.length]

        else:  # unknown content length so parse forever until closed
            while True:
                if self.msg:
                    self.body.extend(self.msg[:])
                    del self.msg[:]  # python2 bytearrays dont have clear self.msg.clear()

                if self.closed:  # no more data so finish
                    break

                (yield None)

        # only gets to here once content length has become finite
        # closed or not chunked or chunking has ended
        self.length = len(self.body)
        self.bodied = True
        (yield True)
        return


class Responder(object):
    """
    Nonblocking HTTP WSGI Responder class

    """
    HttpVersionString = httping.HTTP_11_VERSION_STRING  # http version string

    def __init__(self,
                 incomer,
                 app,
                 environ,
                 chunkable=False):
        """
        Initialize Instance
        Parameters:
            incomer = incomer connection instance
            app = wsgi app callable
            environ = wsgi environment dict
            chunkable = True if may send body in chunks
        """
        self.incomer = incomer
        self.app = app
        self.environ = environ
        self.chunkable = True if chunkable else False

        self.started = False  # True once start called (start_response)
        self.headed = False  # True once headers sent
        self.chunked = False  # True if should send in chunks
        self.ended = False  # True if response body completely sent
        self.closed = False  # True if connection closed by far side
        self.iterator = None  # iterator on application body
        self.status = 200  # integer or string with reason
        self.headers = lodict()  # headers
        self.length = None  # if content-length provided must not exceed
        self.size = 0  # number of body bytes sent so far

    def close(self):
        """
        Close any resources
        """
        self.closed = True

    def reset(self, environ, chunkable=None):
        """
        Reset attributes for another request-response
        """
        self.environ = environ

        if self.chunkable is not None:
            self.chunkable = chunkable

        self.started = False
        self.headed = False
        self.chunked = False
        self.ended = False
        self.iterator = None
        self.status = 200
        self.headers = lodict()
        self.length = None
        self.size = 0

    def build(self):
        """
        Return built head bytes from .status and .headers

        """
        lines = []

        if isinstance(self.status, (int, long)):
            status = "{0} {1}".format(self.status, httping.STATUS_DESCRIPTIONS[self.status])
        else:
            status = self.status

        startLine = "{0} {1}".format(self.HttpVersionString, status)
        try:
            startLine = startLine.encode('ascii')
        except UnicodeEncodeError:
            startLine = startLine.encode('idna')
        lines.append(startLine)

        if u'server' not in self.headers:  # create Server header
            self.headers[u'server'] = "Ioflo WSGI Server"

        if u'date' not in self.headers:  # create Date header
            self.headers[u'date'] = httping.httpDate1123(datetime.datetime.utcnow())

        if self.chunkable and 'transfer-encoding' not in self.headers:
            self.chunked = True
            self.headers[u'transfer-encoding'] = u'chunked'

        for name, value in self.headers.items():
            lines.append(httping.packHeader(name, value))

        lines.extend((b"", b""))
        head = CRLF.join(lines)  # b'/r/n'

        return head

    def write(self, msg):
        """
        WSGI write callback This writes out the headers the first time its called
        otherwise writes the msg bytes
        """
        if not self.started:
            raise AssertionError("WSGI write() before start_response()")

        if not self.headed:  # head not written yet
            head = self.build()
            self.incomer.tx(head)
            self.headed = True

        if self.chunked:
            msg = httping.packChunk(msg)

        if self.length is not None:  # limit total size to length
            size = self.size + len(msg)
            if size > self.length:
                msg = msg[:self.length - size]
            self.size += len(msg)

        self.incomer.tx(msg)

    def start(self, status, response_headers, exc_info=None):
        """
        WSGI application start_response callable

        Parameters:

        status is string of status code and status reason '200 OK'

        response_headers is list of tuples of strings of the form (field, value)
                          one tuple for each header example:
                          [
                              ('Content-type', 'text/plain'),
                              ('X-Some-Header', 'value')
                          ]

        exc_info is optional exception info if exception occurred while
                    processing request in wsgi application
                    If exc_info is supplied, and no HTTP headers have been output yet,
                    start_response should replace the currently-stored
                    HTTP response headers with the newly-supplied ones,
                    thus allowing the application to "change its mind" about
                    the output when an error has occurred.

                    However, if exc_info is provided, and the HTTP headers
                    have already been sent, start_response must raise an error,
                    and should re-raise using the exc_info tuple. That is:

                    raise exc_info[1].with_traceback(exc_info[2]) (python3)
                    raise exc_info[0], exc_info[1], exc_info[2] (python2)
                    Use six.reraise to work for both

        """
        if exc_info:
            try:
                if self.headed:
                    # Re-raise original exception if headers sent
                    reraise(*exc_info)  # sixing.reraise
            finally:
                exc_info = None         # avoid dangling circular ref
        elif self.started:  # may not call start_response again without exc_info
            raise AssertionError("Already started!")

        self.status = status
        self.headers = lodict(response_headers)

        if u'content-length' in self.headers:
            self.length = int(self.headers['content-length'])
            self.chunkable = False  # cannot use chunking with finite content-length
        else:
            self.length = None
            self.chunkable = self.chunkable

        self.started = True
        return self.write

    def service(self):
        """
        Service application
        """
        if not self.closed and not self.ended:
            if self.iterator is None:  # initiate application
                self.iterator = iter(self.app(self.environ, self.start))

            try:
                msg = next(self.iterator)
            except StopIteration:
                self.ended = True
            else:
                self.write(msg)
                if self.length is not None and self.size >= self.length:
                    self.ended = True


class Valet(object):
    """
    Valet WSGI Server Class
    """
    Timeout = 5.0  # default server connection timeout

    def __init__(self,
                 app=None,
                 reqs=None,
                 reps=None,
                 servant=None,
                 store=None,
                 name='',
                 bufsize=8096,
                 wlog=None,
                 ha=None,
                 host=u'',
                 port=None,
                 eha=None,
                 scheme=u'',
                 timeout=None,
                 **kwa):
        """
        Initialization method for instance.
        app = wsgi application callable
        reqs = odict of Requestant instances keyed by ca
        apps = odict of running WsgiResponder instances keyed by ca
        servant = instance of Server or ServerTls or None
        store = Datastore for timers

        kwa needed to pass additional parameters to servant

        if servantinstances are not provided (None)
        some or all of these parameters will be used for initialization

        name = user friendly name for servant
        bufsize = buffer size
        wlog = WireLog instance if any
        ha = host address duple (host, port) for local servant listen socket
        host = host address for local servant listen socket, '' means any interface on host
        port = socket port for local servant listen socket
        eha = external destination address for incoming connections used in TLS
        scheme = http scheme u'http' or u'https' or empty

        """
        self.app = app
        self.reqs = reqs if reqs is not None else odict()
        self.reps = reps if reps is not None else odict()
        self.store = store or storing.Store(stamp=0.0)
        if not name:
            name = "Ioflo_WSGI_server"
        self.timeout = timeout if timeout is not None else self.Timeout

        ha = ha or (host, port)  # ha = host address takes precendence over host, port
        if servant:
            if isinstance(servant, ServerTls):
                if scheme and scheme != u'https':
                    raise  ValueError("Provided scheme '{0}' incompatible with servant".format(scheme))
                secured = True
                scheme = u'https'
                defaultPort = 443
            elif isinstance(servant, Server):
                if scheme and scheme != u'http':
                    raise  ValueError("Provided scheme '{0}' incompatible with servant".format(scheme))
                secured = False
                scheme = 'http'
                defaultPort = 80
            else:
                raise ValueError("Invalid servant type {0}".format(type(servant)))
        else:
            scheme = u'https' if scheme == u'https' else u'http'
            if scheme == u'https':
                secured = True  # use tls socket connection
                defaultPort = 443
            else:
                secured = False # non tls socket connection
                defaultPort = 80

        self.scheme = scheme
        host, port = ha
        port = port or  defaultPort  # if port not specified
        ha = (host, port)

        if servant:
            if servant.ha != ha:
                ValueError("Provided ha '{0}:{1}' incompatible with servant".format(ha[0], ha[1]))
            # at some point may want to support changing the ha of provided servant

        else:  # what about timeouts for servant connections
            if secured:
                servant = ServerTls(store=self.store,
                                    name=name,
                                    ha=ha,
                                    eha=eha,
                                    bufsize=bufsize,
                                    wlog=wlog,
                                    timeout=self.timeout,
                                    **kwa)
            else:
                servant = Server(store=self.store,
                                 name=name,
                                 ha=ha,
                                 eha=eha,
                                 bufsize=bufsize,
                                 wlog=wlog,
                                 timeout=self.timeout,
                                 **kwa)


        self.secured = secured
        self.servant = servant

    def idle(self):
        """
        Returns True if no connections have requests in process
        Useful for debugging
        """
        idle = True
        for requestant in self.reqs.values():
            if not requestant.ended:
                idle = False
                break
        if idle:
            for responder in self.reps.values():
                if not responder.ended:
                    idle = False
                    break
        return idle

    def refreshIx(self, ca):
        """
        Restart incomer timer given by ca
        """
        if ca in self.ixes:
            self.ixes[ca].timer.restart()

    def buildEnviron(self, requestant):
        """
        Returns wisgi environment dictionary for supplied requestant
        """
        environ = odict()  # maybe should be modict for cookies or other repeated headers

        # WSGI variables
        environ['wsgi.version'] = (1, 0)
        environ['wsgi.url_scheme'] = self.scheme
        environ['wsgi.input'] = io.BytesIO(requestant.body)
        environ['wsgi.errors'] = sys.stderr
        environ['wsgi.multithread'] = False
        environ['wsgi.multiprocess'] = False
        environ['wsgi.run_once'] = False
        environ["wsgi.server_name"] = self.servant.name
        environ["wsgi.server_version"] = (1, 0)

        # Required CGI variables
        environ['REQUEST_METHOD'] = requestant.method      # GET
        environ['SERVER_NAME'] = self.servant.eha[0]      # localhost
        environ['SERVER_PORT'] = str(self.servant.eha[1])  # 8888
        environ['SERVER_PROTOCOL'] = "HTTP/{0}.{1}".format(*requestant.version)  # used by request http/1.1
        environ['SCRIPT_NAME'] = u''
        environ['PATH_INFO'] = requestant.path        # /hello?name=john

        # Optional CGI variables
        environ['QUERY_STRING'] = requestant.query        # name=john
        environ['REMOTE_ADDR'] = requestant.incomer.ca
        environ['CONTENT_TYPE'] = requestant.headers.get('content-type', '')
        length = requestant.length if requestant.length is not None else ""
        environ['CONTENT_LENGTH'] = length

        # recieved http headers mapped to all caps with HTTP_ prepended
        for key, value in requestant.headers.items():
            key = "HTTP_" + key.replace("-", "_").upper()
            environ[key] = value

        return environ

    def closeConnection(self, ca):
        """
        Close and remove connection and associated steward given by ca
        """
        self.servant.removeIx(ca)
        self.reqs[ca].close()
        del self.reqs[ca]
        self.reps[ca].close()
        del self.reps[ca]

    def serviceConnects(self):
        """
        Service new incoming connections
        Create requestants
        Timeout stale connections
        """
        self.servant.serviceConnects()
        for ca, ix in self.servant.ixes.items():
            if ix.cutoff:
                if ca in self.reqs:
                    self.reqs[ca].close()  # this signals parser
                if ca in self.reps:
                    self.reps[ca].close()  # this signals handler
                self.closeConnection(ca)
                continue

            if ca not in self.reqs:
                self.reqs[ca] = Requestant(msg=ix.rxbs, incomer=ix)

            if ix.timeout > 0.0 and ix.timer.expired:
                self.closeConnection(ca)

    def serviceReqs(self):
        """
        Service pending requestants
        """
        for ca, requestant in self.reqs.items():
            if requestant.parser:
                try:
                    requestant.parse()
                except httping.HTTPException as ex:
                    #requestant.errored = True
                    #requestant.error = str(ex)
                    #requestant.ended = True
                    sys.stderr.write(str(ex))
                    self.closeConnection(ca)
                    continue

                if requestant.changed:
                    self.refreshIx(ca)

                if requestant.ended:
                    console.concise("Parsed Request:\n{0} {1} {2}\n"
                                    "{3}\n{4}\n".format(requestant.method,
                                                        requestant.path,
                                                        requestant.version,
                                                        requestant.headers,
                                                        requestant.body))
                    # create or restart wsgi app responder here
                    environ = self.buildEnviron(requestant)
                    if ca not in self.reps:
                        chunkable = True if requestant.version >= (1, 1) else False
                        responder = Responder(incomer=requestant.incomer,
                                                  app=self.app,
                                                  environ=environ,
                                                  chunkable=chunkable)
                        self.reps[ca] = responder
                    else:  # reuse
                        responder = self.reps[ca]
                        responder.reset(environ=environ)

    def serviceReps(self):
        """
        Service pending responders
        """
        for ca, respondent in self.reps.items():
            if respondent.closed:
                self.closeConnection(ca)
                continue

            if not respondent.ended:
                respondent.service()

            if respondent.ended:
                requestant = self.reqs[ca]
                if requestant.persisted:
                    if requestant.parser is None:  # reuse
                        requestant.makeParser()  # resets requestant parser
                else:  # not persistent so close and remove requestant and responder
                    self.closeConnection(ca)

    def serviceAll(self):
        """
        Service request response
        """
        self.serviceConnects()
        self.servant.serviceReceivesAllIx()
        self.serviceReqs()
        self.serviceReps()
        self.servant.serviceTxesAllIx()


class CustomResponder(object):
    """
    Nonblocking HTTP Server Response class

    HTTP/1.1 200 OK\r\n
    Content-Length: 122\r\n
    Content-Type: application/json\r\n
    Date: Thu, 30 Apr 2015 19:37:17 GMT\r\n
    Server: IoBook.local\r\n\r\n
    """
    HttpVersionString = httping.HTTP_11_VERSION_STRING  # http version string

    def __init__(self,
                 steward=None,
                 status=200,  # integer
                 headers=None,
                 body=b'',
                 data=None):
        """
        Initialize Instance
        steward = managing Steward instance
        status = response status code
        headers = http response headers
        body = http response body
        data = dict to jsonify as body if provided
        """
        self.steward = steward
        self.status = status
        self.headers = lodict(headers) if headers else lodict()
        if body and isinstance(body, unicode):  # use default
            # RFC 2616 Section 3.7.1 default charset of iso-8859-1.
            body = body.encode('iso-8859-1')
        self.body = body or b''
        self.data = data

        self.ended = False  # True if response generated completed

        self.msg = b""  # for debugging
        self.lines = []  # for debugging
        self.head = b""  # for debugging


    def reinit(self,
               status=None,  # integer
               headers=None,
               body=None,
               data=None):
        """
        Reinitialize anything that is not None
        This enables creating another response on a connection
        """
        if status is not None:
            self.status = status
        if headers is not None:
            self.headers = lodict(headers)
        if body is not None:  # body should be bytes
            if isinstance(body, unicode):
                # RFC 2616 Section 3.7.1 default charset of iso-8859-1.
                body = body.encode('iso-8859-1')
            self.body = body
        else:
            self.body = b''
        if data is not None:
            self.data = data
        else:
            self.data = None

    def build(self,
              status=None,
              headers=None,
              body=None,
              data=None):
        """
        Build and return response message

        """
        self.reinit(status=status,
                    headers=headers,
                    body=body,
                    data=data)
        self.lines = []

        startLine = "{0} {1} {2}".format(self.HttpVersionString,
                                         self.status,
                                         httping.STATUS_DESCRIPTIONS[self.status])
        try:
            startLine = startLine.encode('ascii')
        except UnicodeEncodeError:
            startLine = startLine.encode('idna')
        self.lines.append(startLine)

        if u'server' not in self.headers:  # create Server header
            self.headers[u'server'] = "Ioflo Server"

        if u'date' not in self.headers:  # create Date header
            self.headers[u'date'] = httping.httpDate1123(datetime.datetime.utcnow())

        if self.data is not None:
            body = ns2b(json.dumps(self.data, separators=(',', ':')))
            self.headers[u'content-type'] = u'application/json; charset=utf-8'
        else:
            body = self.body

        if body and (u'content-length' not in self.headers):
            self.headers[u'content-length'] = str(len(body))

        for name, value in self.headers.items():
            self.lines.append(httping.packHeader(name, value))

        self.lines.extend((b"", b""))
        self.head = CRLF.join(self.lines)  # b'/r/n'

        self.msg = self.head + body
        self.ended = True
        return self.msg

class Steward(object):
    """
    Manages the associated requestant and responder for an incoming connection
    """
    def __init__(self,
                 incomer,
                 requestant=None,
                 responder=None,
                 dictable=False):
        """
        incomer = Incomer instance for connection
        requestant = Requestant instance for connection
        responder = Responder instance for connection
        dictable = True if should attempt to convert request body as json
        """
        self.incomer = incomer
        if requestant is None:
            requestant = Requestant(msg=self.incomer.rxbs,
                                    incomer=incomer,
                                    dictable=dictable)
        self.requestant = requestant

        if responder is None:
            responder = CustomResponder(steward=self)
        self.responder = responder
        self.waited = False  # True if waiting for reponse to finish
        self.msg = b""  # outgoing msg bytes

    def refresh(self):
        """
        Restart incomer timer
        """
        incomer.timer.restart()

    def respond(self):
        """
        Respond to request  Override in subclass
        Echo request
        """
        console.concise("Responding to Request:\n{0} {1} {2}\n"
                                "{3}\n{4}\n".format(self.requestant.method,
                                                    self.requestant.path,
                                                    self.requestant.version,
                                                    self.requestant.headers,
                                                    self.requestant.body))
        data = odict()
        data['version'] = "HTTP/{0}.{1}".format(*self.requestant.version)
        data['method'] = self.requestant.method

        pathSplits = urlsplit(unquote(self.requestant.url))
        path = pathSplits.path
        data['path'] = path

        query = pathSplits.query
        qargs = odict()
        qargs, query = httping.updateQargsQuery(qargs, query)
        data['qargs'] = qargs

        fragment = pathSplits.fragment
        data['fragment'] = fragment

        data['headers'] = copy.copy(self.requestant.headers)  # make copy
        data['body'] = self.requestant.body.decode('utf-8')
        data['data'] = copy.copy(self.requestant.data)  # make copy

        msg = self.responder.build(status=200, data=data)
        self.incomer.tx(msg)
        self.waited = not self.responder.ended

    def pour(self):
        """
        Run generator to stream response message

        """

        # putnext generator here

        if self.responder.ended:
            self.waited = False
        else:
            self.refresh()


class Porter(object):
    """
    Porter class nonblocking HTTP server
    """
    Timeout = 5.0  # default server connection timeout

    def __init__(self,
                 servant=None,
                 store=None,
                 stewards=None,
                 name='',
                 bufsize=8096,
                 wlog=None,
                 ha=None,
                 host=u'',
                 port=None,
                 eha=None,
                 scheme=u'',
                 dictable=False,
                 timeout=None,
                 **kwa):
        """
        Initialization method for instance.
        servant = instance of Server or ServerTls or None
        store = Datastore for timers
        stewards = odict of Steward instances
        kwa needed to pass additional parameters to servant

        if servantinstances are not provided (None)
        some or all of these parameters will be used for initialization

        name = user friendly name for servant
        bufsize = buffer size
        wlog = WireLog instance if any
        ha = host address duple (host, port) for local servant listen socket
        host = host address for local servant listen socket, '' means any interface on host
        port = socket port for local servant listen socket
        eha = external destination address for incoming connections used in TLS
        scheme = http scheme u'http' or u'https' or empty
        dictable = Boolean flag If True attempt to convert body from json for requestants

        """
        self.store = store or storing.Store(stamp=0.0)
        self.stewards = stewards if stewards is not None else odict()
        self.dictable = True if dictable else False  # for stewards
        self.timeout = timeout if timeout is not None else self.Timeout

        ha = ha or (host, port)  # ha = host address takes precendence over host, port
        if servant:
            if isinstance(servant, ServerTls):
                if scheme and scheme != u'https':
                    raise  ValueError("Provided scheme '{0}' incompatible with servant".format(scheme))
                secured = True
                scheme = u'https'
                defaultPort = 443
            elif isinstance(servant, Server):
                if scheme and scheme != u'http':
                    raise  ValueError("Provided scheme '{0}' incompatible with servant".format(scheme))
                secured = False
                scheme = 'http'
                defaultPort = 80
            else:
                raise ValueError("Invalid servant type {0}".format(type(servant)))
        else:
            scheme = u'https' if scheme == u'https' else u'http'
            if scheme == u'https':
                secured = True  # use tls socket connection
                defaultPort = 443
            else:
                secured = False # non tls socket connection
                defaultPort = 80

        host, port = ha
        port = port or  defaultPort  # if port not specified
        ha = (host, port)

        if servant:
            if servant.ha != ha:
                ValueError("Provided ha '{0}:{1}' incompatible with servant".format(ha[0], ha[1]))
            # at some point may want to support changing the ha of provided servant

        else:  # what about timeouts for servant connections
            if secured:
                servant = ServerTls(store=self.store,
                                    name=name,
                                    ha=ha,
                                    eha=eha,
                                    bufsize=bufsize,
                                    wlog=wlog,
                                    timeout=self.timeout,
                                    **kwa)
            else:
                servant = Server(store=self.store,
                                 name=name,
                                 ha=ha,
                                 eha=eha,
                                 bufsize=bufsize,
                                 wlog=wlog,
                                 timeout=self.timeout,
                                 **kwa)


        self.secured = secured
        self.servant = servant

    def idle(self):
        """
        Returns True if no connections have requests in process
        Useful for debugging
        """
        idle = True
        for steward in self.stewards.values():
            if not steward.requestant.ended:
                idle = False
                break
        return idle

    def closeConnection(self, ca):
        """
        Close and remove connection and associated steward given by ca
        """
        self.servant.removeIx(ca)
        del self.stewards[ca]

    def serviceConnects(self):
        """
        Service new incoming connections
        Create requestants
        Timeout stale connections
        """
        self.servant.serviceConnects()
        for ca, ix in self.servant.ixes.items():
            # check for and handle cutoff connections by client here

            if ca not in self.stewards:
                self.stewards[ca] = Steward(incomer=ix, dictable=self.dictable)

            if ix.timeout > 0.0 and ix.timer.expired:
                self.closeConnection(ca)

    def serviceStewards(self):
        """
        Service pending requestants and responders
        """
        for ca, steward in self.stewards.items():
            if not steward.waited:
                steward.requestant.parse()

                if steward.requestant.changed:
                    steward.refresh()

                if steward.requestant.ended:
                    steward.requestant.dictify()
                    console.concise("Parsed Request:\n{0} {1} {2}\n"
                                    "{3}\n{4}\n".format(steward.requestant.method,
                                                        steward.requestant.path,
                                                        steward.requestant.version,
                                                        steward.requestant.headers,
                                                        steward.requestant.body))
                    steward.respond()

            if steward.waited:
                steward.pour()

            if not steward.waited and steward.requestant.ended:
                if steward.requestant.persisted:
                    steward.requestant.makeParser()  #set up for next time
                else:  # remove and close connection
                    self.closeConnection(ca)

    def serviceAll(self):
        """
        Service request response
        """
        self.serviceConnects()
        self.servant.serviceReceivesAllIx()
        self.serviceStewards()
        self.servant.serviceTxesAllIx()

